#include "n64adapter.h"

extern "C" {
#include <avr/io.h>		// AVR port definition
#include <avr/interrupt.h>	// sei()
#include <util/delay.h>		// for _delay_ms()
#include "usbdrv.h"
}
#include "hiddescriptor.h"
#include "n64controller.h"

void sendControllerReport(N64Controller * controller) {
	controller->updateReport();
	switch (controller->getStatus()) {
	case 0:
		PORTB = PORTB | controller->getId();
		break;
	case 1:
		while (!usbInterruptIsReady())
			usbPoll();
		report_t report = controller->getReport();
		usbSetInterrupt(reinterpret_cast<unsigned char*>(&report), sizeof(report));
		PORTB = PORTB | controller->getId();
		break;
	default:
		PORTB = PORTB & ~controller->getId();
		break;
	}
}

void N64Adapter::init() {
	usbInit();
	// Enforce re-enumeration, do this while interrupts are disabled!
	usbDeviceDisconnect();
	uchar i = 0;
	// Fake USB disconnect for > 250 ms
	while(--i){
		_delay_ms(1);
	}
	usbDeviceConnect();
	sei();

	while (true) {
		usbPoll();
		sendControllerReport(&controller1);
		sendControllerReport(&controller2);
		sendControllerReport(&controller3);
		sendControllerReport(&controller4);
	}
}

N64Adapter::N64Adapter() :
	controller1(0x01),
	controller2(0x02),
	controller3(0x04),
	controller4(0x08)
{
	DDRC	= 0x00;
	PORTC	= 0x00;
	DDRB	= 0xff;
};

usbMsgLen_t usbFunctionSetup(uchar data[8])
{
	usbRequest_t *rq = reinterpret_cast<usbRequest_t *>(data);

	if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){ // Class request type
		if(rq->bRequest == USBRQ_HID_GET_REPORT){ // wValue: ReportType (highbyte), ReportID (lowbyte)
			// Get the report ID
			return 0;
		}
	}
	return 0; // Default for non implemented definitions
}
