#include "n64controller.h"

extern "C" {
#include <avr/io.h>
#include <util/delay.h>
}

#define N64_DATA_DDR DDRC
#define N64_DATA_PIN PINC

N64Controller::N64Controller(unsigned char id) : id(id), report()
{
	report.controllerNumber = id;
};

void N64Controller::updateReport()
{
	// Create an array to keep the button's status information
	unsigned char volatile results[4];
	unsigned char count = 0;
	unsigned char request = 0x01; // get controller status
	/*
	 * z: Points to current source byte
	 *
	 * 3 us == 48 cycles
	 * 1 us == 16 cycles
	 */

	asm volatile(
		"	push r30		\n"
		"	push r31		\n"
		"	ldi r19, 0x0		\n" // 1
		"	out %1, r19		\n" // 1
		"	ldi r16, 0x80		\n" // 1
		"nextBit:			\n"
		"	mov r17, %2		\n" // 1
		"	and r17, r16		\n" // 1
		"	breq send0		\n" // 2   4
		"	nop			\n"
		// Send a 1: 1us Low, 3us High
		"send1:				\n"
		"	nop			\n" // 1
		"	out %1, %5		\n" // 1
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop		\n" // 2
		"	ldi r19, 0x0		\n" // 1
		"	out %1, r19		\n" // 1
		"	ldi r19, 13		\n" // 1
		"lp1:				\n"
		"	dec r19			\n" // 1
		"	brne lp1		\n" // 2
		"	nop\nnop		\n" // 2
		"	lsr r16			\n" // 1
		"	breq done		\n" // 1
		"	rjmp nextBit		\n" // 2
		// Send a 0: 3us Low, 1us High
		"send0:				\n"
		"	nop			\n" // 1
		"	out %1, %5		\n" // 1
		"	ldi r19, 16		\n" // 1
		"lp0:				\n"
		"	dec r19			\n" // 1
		"	brne lp0		\n" // 2
		"	nop			\n" // 1
		"	ldi r19, 0x0		\n" // 1
		"	out %1, r19		\n" // 1
		"	nop\nnop		\n" // 2
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	lsr r16			\n" // 1
		"	breq done		\n" // 1
		"	rjmp nextBit		\n" // 2
		"done:				\n"
		"	ldi r19, 0x0		\n" // 1
		"	out %1, r19		\n" // 1
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop			\n" // 1
		// Stop bit (1us low, 3us high)
		"	nop			\n" // 1
		"	out %1, %5		\n" // 1
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop		\n" // 2
		"	ldi r19, 0x0		\n" // 1
		"	out %1, r19		\n" // 1

		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop		\n" // 3
											// 47 ciclos
		// Response reading part //
		// r16: temp1
		// r17: temp2
		// r18: temp3
		// r19: temp4
		// r20: bit counter
		// r21: bit buffer
		// r22: pointer backup
		"	clr r20			\n" // 1 clear bit counter
											// 48 ciclos, stop bit done
		// OK, so there is now a 0 on the wire.
		// Worst case, we are at the 9th cycle.
		// Best case, we are at the 4th cycle.
		// 	Middle: cycle 6
		//
		// cycle: 1-16 16-32 32-48 48-64
		//	high:  0     1     1     1
		//	low:   0     0     0     1
		//
		// I check the pin on the 32nd cycle which is the safest place.
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop\nnop	\n" // 4
		"	nop\nnop\nnop		\n" // 3
		"	mov r22, r30		\n" // 1 cycle 32
		// We are now more or less aligned on the 32nd cycle.
		"inData:			\n"
		"	in r18, %4		\n" // 1  Read from the port
		"	mov r16, %5		\n" // 1  
		"	and r18, %5		\n" // 1  Isolate our bit - controller 1
		"	lsl	r21		\n" // 1  shif byte register controller 1
		"shift:				\n"
		"	lsr r16			\n" // 1
		"	breq done1		\n" // 1
		"	lsr r18			\n" // 1
		"	rjmp shift		\n" // 2
		"done1:	or r21, r18		\n" // 1
		"	st z, r21		\n" // 2  store the value
		// Wait for a high level at the wire
		"	ldi r16, 0x10		\n" // setup a timeout
		"waitHigh:			\n"
		"	dec r16			\n" // decrement timeout
		"	breq timeout1		\n" // handle timeout condition
		"	in r18, %4		\n" // Read the port
		"	and r18, %5		\n" // Isolate our bit
		"	breq waitHigh		\n" // Continue if set
		// Wait for a low level at the wire
		"	ldi r16, 0x10		\n" // setup a timeout
		"waitFall:			\n"
		"	dec r16			\n" // 1
		"	breq timeout		\n" // 1
		"	in r17, %4		\n" // 1  read the input port
		"	and r17, %5		\n" // 1  isolate the input bit
		"	brne waitFall		\n" // 2  if still high, loop
		"	inc r20			\n" // 1 incrementa contador de bits
		"	mov r19, r20		\n" // 1 copia o contador de bits
		"	lsr r19			\n" // 1
		"	lsr r19			\n" // 1
		"	lsr r19			\n" // 1 divide per 8
		"	mov r30, r22		\n" // 1 compensate previous sum
		"	add r30, r19		\n" // 1 add new offset
		"	mov r19, r20		\n" // 1 copy bit counter
		"	subi r19, 32		\n" // 1 see if we got 32 bits
		"	breq ok			\n" // 1 branch if positive
		"	nop			\n" // 1
		"	nop\nnop\nnop\nnop	\n" // 4
		"	rjmp inData		\n" // 2 otherwise read next bit
		"ok:				\n" // Everything allright
		"	ldi %0, 32		\n"
		"	rjmp end		\n"
		"timeout1:			\n" // timeout waiting high level
		"	ldi %0, 2		\n"
		"	rjmp end		\n"
		"timeout:			\n" // timeout waiting low level
		"	ldi %0, 1		\n"
		"	rjmp end		\n"
		"end:				\n"
		"	pop r31			\n"
		"	pop r30			\n"
		: "=r" (count)
		: "I" (_SFR_IO_ADDR(N64_DATA_DDR)), "r"(request),
			"z"(results), "I" (_SFR_IO_ADDR(N64_DATA_PIN)),
			"r" (id)
		: "r16","r17","r18","r19","r20","r21","r22"
	);

	status = 0;

	if (count != 32){
		_delay_us(128);
		status =  2;
		return;
	}

	if (((int) ((signed char) results[2])) > 71 )
		results[2] = 255;
	else if (((int) ((signed char) results[2])) <= -71 )
		results[2] = 0;
	else
		results[2] = (((int) ((signed char) results[2]))*3)/2+127;

	if (((int) ((signed char) results[3])) > 71 )
		results[3] = 255;
	else if (((int) ((signed char) results[3])) <= -71 )
		results[3] = 0;
	else
		results[3] = (((int) ((signed char) results[3]))*3)/2+127;

	results[3] = ~results[3]; // Invert axis

	// Check if buttonsL has changed
	if (report.buttonsL != results[0]){
		report.buttonsL = results[0];
		status = 1;
	}

	// Check if buttonsH has changed
	if (report.buttonsH != results[1]){
		report.buttonsH = results[1];
		status = 1;
	}

	// Check if directional X has changed
	if (report.dx != results[2]){
		report.dx = results[2];
		status = 1;
	}

	// Check if directional Y has changed
	if (report.dy != results[3]){
		report.dy = results[3];
		status = 1;
	}
};
